<?php

//! A manager for one's own notification subscriptions or for subscriptions of other users.
function comments_notifications_manager(): string
{
	global $comments_config,$currentPage;
	$user=$comments_config['user'];
	$userId=$user['id'];

	if(isset($_GET['userId']) && $_GET['userId']!=$user['id'])
	{
		if(!$user['isSuperAdmin']){dlib_permission_denied_403();}
		$userId=intval($_GET['userId']);
	}
	else
	if(!$user['isAdmin']){dlib_permission_denied_403();}

	$currentPage->addCss('dlib/comments/comments-manager.css');

	$currentPage->addJsTranslations(['subscribe'            =>t('Subscribe'),
									 're-subscribe'         =>t('Re-subscribe'),
									 'cleanup-subscribtion' =>t('Cleaned up old subscription.'),
									]);

	$notificationsVapidKeys=comments_notifications_get_vapid_keys();   
	$currentPage->addJsVar('notifications_vapid_public_key',$notificationsVapidKeys['publicKey']);

	$currentPage->addJs('lib/jquery.js');
	$currentPage->addJs('dlib/comments/comments-notifications-manager.js');

	$currentPage->addJsVar('isOwnNotifications',$userId==$user['id']);

	$subscriptions=variable_get('notifications-subscription','',[]);
	$subscriptions=array_filter($subscriptions,function($s)use($userId){return $s['userId']==$userId;});

	return template_render('comments-notifications-manager.tpl.php',
		   [compact('comments_config','user','userId','subscriptions')]);
}

//! Returns saved VAPID keys or creates them if this is the first time it is called.
//! VAPID keys are used by notifications to identify our service.
function comments_notifications_get_vapid_keys(): array
{
	$res=variable_get('notifications_vapid_keys');
	if($res!==false){return $res;}

	// Generating VAPID keys with openssl is not (yet?) documented. 
	// Probably some issues with key formats, padding, base64 variants....
	// We use our web-push proxy to generate them : comments_notifications_web_push_proxy("generate-vapid-keys")
	//$privKeyR=openssl_pkey_new(["private_key_type"=>OPENSSL_KEYTYPE_EC,"curve_name"=>"prime256v1"]);
	//$details=openssl_pkey_get_details($privKeyR);
	//$pubKey=$details['key'];
	//$privKey=str_replace(['=','+','/'],['','-','_'],base64_encode($details['ec']['d']));
	$json=comments_notifications_web_push_proxy("generate-vapid-keys");
	$keys=json_decode($json,true);
	if($keys===null)
	{
		dlib_log_error('generate-vapid-keys failed');
		$res=['publicKey'=>false,'privateKey'=>false];
	}
	else
	{
		$res=['publicKey'=>$keys['publicKey'],'privateKey'=>$keys['privateKey']];
		variable_set('notifications_vapid_keys','',$res);
	}
	return $res;
}

//! Ajax request to subscribe the current user to notifications
function comments_notifications_subscription_save(): array
{
	global $comments_config,$currentPage;
	$user=$comments_config['user'];
	if(!$user['isAdmin']){dlib_permission_denied_403();}

	// validation: just some sanity checks
	if(!isset($_POST['subscription']) || 
	   !isset($_POST['subscription']['endpoint']) || 
	   !isset($_POST['subscription']['keys']) || 
	   !isset($_POST['subscription']['keys']['auth']) || 
	   !isset($_POST['subscription']['keys']['p256dh'])){dlib_bad_request_400();}

	if(!preg_match('@^https://.{20,2000}$@'   ,$_POST['subscription']['endpoint']      )){dlib_bad_request_400();}
	if(!preg_match('@^[a-zA-Z0-9_-]{5,200}$@' ,$_POST['subscription']['keys']['auth'  ])){dlib_bad_request_400();}
	if(!preg_match('@^[a-zA-Z0-9_-]{5,1000}$@',$_POST['subscription']['keys']['p256dh'])){dlib_bad_request_400();}

	require_once "lib/UserAgentParser.php";

	$endpoint=$_POST['subscription']['endpoint'];
	$subscriptions=variable_get('notifications-subscription','',[]);
	$subscriptions[$endpoint]=
		[
			'userId'    =>$user['id'],
			'time'      =>time(),
			'userAgent' =>comments_user_agent_format(null,false),
			'lastSent'  =>[],
			'nightStart'=>'23:00',
			'nightEnd'  =>'08:00',
			'subscription'=>
			[
				//! endpoint is a URL generated by the browser that identifies a single browser user profile
				'endpoint'=>$endpoint,
				'keys'=>
					[
						'auth'      =>$_POST['subscription']['keys']['auth'],
						'p256dh'    =>$_POST['subscription']['keys']['p256dh'],
					],
			],
		];
	variable_set('notifications-subscription','',$subscriptions);
	return ['status'=>'ok'];
}

//! Ajax request to delete notification subscription
function comments_notifications_subscription_update(): array
{
	global $comments_config,$currentPage;
	$res=[];
	$user=$comments_config['user'];
	// Note: we don't need to check for isSuperAdmin when updating somebody else's subscriptions, as only superAdmin can know secret endpoint url.
	if(!$user['isAdmin']){dlib_permission_denied_403();}

	if(!isset($_POST['endpoint'])){dlib_bad_request_400();}
	$endpoint=$_POST['endpoint'];

	$subscriptions=variable_get('notifications-subscription','',[]);
	if(!isset($subscriptions[$endpoint])){dlib_not_found_404();}

	switch($_POST['action'])
	{
		case 'delete':
			unset($subscriptions[$endpoint]);
			$res=['status'=>'ok'];
			break;
		case 'night':
			$nightStart=$_POST['nightStart'];
			$nightEnd  =$_POST['nightEnd'  ];
			if(!($nightStart==='' && $nightEnd==='') && (
			   !preg_match('@^(0[0-9]|1[0-9]|2[0-3]):[0-5][0-9]$@',$nightStart) ||
			   !preg_match('@^(0[0-9]|1[0-9]|2[0-3]):[0-5][0-9]$@',$nightEnd  )   )){dlib_bad_request_400();}
			$subscriptions[$endpoint]['nightStart']=$nightStart;
			$subscriptions[$endpoint]['nightEnd'  ]=$nightEnd;
			$res=['status'=>'ok'];
			break;
	default: dlib_bad_request_400();
	}

	variable_set('notifications-subscription','',$subscriptions);
	return $res;
}

//! Send Web Push notifications about this comment to all subscribed browsers ("endpoints").
function comments_notifications_send_all(Comment $comment)
{
	global $comments_config;
	$subscriptions=variable_get('notifications-subscription','',[]);
	$now=time();
	$delete=[];
	foreach($subscriptions as $endpoint=>&$subscription)
	{
		// Do not send notifications at night
		if(($subscription['nightStart'] ?? '')!=='')
		{
			$todayStart=strtotime('today '.$subscription['nightStart']);
			$todayEnd  =strtotime('today '.$subscription['nightEnd'  ]);
			if($todayStart>$todayEnd && ( time()<$todayEnd || time()>$todayStart ) ){continue;}
			if($todayStart<$todayEnd && ( time()<$todayEnd && time()>$todayStart ) ){continue;}
		}

		// FIXMEc: should we check for multiple sends in less than x time ?
		$payload=json_encode(
			[
				'title'=>t('Comment').':'.($comment->isPublished ? 'P' : 'N').':'.$comment->blurb(),
				'body' =>$comment->userName().' posted a comment', // FIXMEc: find better info. Check if this needs to be html safe 
				'data' =>['url'=>$comments_config['link']($comment)],
			]);
		$ret=comments_notifications_send($subscription,$payload);
		if($ret==='delete'){$delete[]=$endpoint;}
		else
		if($ret==='ok')
		{
			$subscription['lastSent'][]=['time'=>$now,'type'=>'comment','id'=>$comment->id];
			// just in case something crashes before the end of this function
			variable_set('notifications-subscription','',$subscriptions);
		}
		else
		{
			dlib_log_error('comments_notifications_send failed for userId='.$subscription['userId']."\n".$endpoint);
		}
	}
	if(count($delete))
	{
		foreach($delete as $endpoint){unset($subscriptions[$endpoint]);}
	}
	variable_set('notifications-subscription','',$subscriptions);
}

//! Send a notification to a single user ($subscription / "endpoint")
function comments_notifications_send(array $subscription,string $payload): string
{
	global $comments_config;
	$notificationsVapidKeys=variable_get('notifications_vapid_keys');
	if($notificationsVapidKeys===false){return false;}
	$res=false;
	$args=
		[
			'subscription'=>$subscription['subscription'],
			'payload'=>$payload,
			'options'=>
			[
				'vapidDetails'=>
				[
					'subject'=>$comments_config['notifications_vapid_subject'],
					'publicKey' =>$notificationsVapidKeys['publicKey'],
					'privateKey'=>$notificationsVapidKeys['privateKey'],
				],
				'TTL'=>3600*24*2,
			],
		];
	$resultRaw=comments_notifications_web_push_proxy('send-notification',$args);
	if($resultRaw===false){dlib_log_error('comments_notifications_web_push_proxy failed, check general log');return false;}
	$result=json_decode($resultRaw,true);
	//vd('result',$resultRaw,$result);
	if($result['statusCode']>=400 && ($result['name'] ?? '')==='WebPushError')
	{
		$error=json_decode($result['body'],true);
		// Remove permanently invalid endpoints
		// https://autopush.readthedocs.io/en/latest/http.html
		if($result['statusCode']===404 || $result['statusCode']===410){$res='delete';}
	}
	if($result['statusCode']<400){$res='ok';}
	return $res;
}

//! Send command to proxy that calls the Node-js "web-push" command
function comments_notifications_web_push_proxy(string $command,array $args=[])
{
	// Create socket
	$socket=@socket_create(AF_INET, SOCK_STREAM, SOL_TCP);
	if($socket===false)
	{
		dlib_log('web-push : socket_create() failed: reason: ' . socket_strerror(socket_last_error()));
		return false;
	}

	$result=@socket_connect($socket,'127.0.0.1',5362);
	if ($result===false)
	{
		dlib_log('web-push : socket_connect() failed ('.$result.') : ' . socket_strerror(socket_last_error($socket)));
		return false;
	}

	$data=json_encode(['command'=>$command,'args'=>$args])."\r\n";
	$l=0;
	for($l=0;$l<strlen($data);$l+=$written)
	{
		$written=socket_write($socket, $data, strlen($data));
		if($written===false)
		{
			dlib_log('web-push write failed');
			return false;
		}
	}

	// Server sends us a single JSON line + newline to tell us it has finished.
	$read=socket_read($socket, 100000, PHP_NORMAL_READ);

	return $read;
}
